DRWATERGITROOT <- "git.drwater.net"
DRWATERGITSCHEME <- "https"
GITROOTURL <- paste0(DRWATERGITSCHEME, "://", DRWATERGITROOT)
DRWATERPKGS <- c(
  "dwfun", "langeR", "rmdify", "dateR", "dfeR",
  "limvis", "uniteR", "cctda", "cctdb",
  "mapeR", "meteR", "drwateR", "uniteR", "dataSHR",
  "yuqiao"
)
# Define core packages
CORE <- c(
  "dwfun", "langeR", "rmdify", "dateR",
  "dfeR", "uniteR", "drwateR"
)
PACKAGETYPE <- c("public", "protected")
# Define the repository URL
DRANREPOURL <- "https://dran.drwater.net"
# Path to the .Rprofile file
RPROFILEFN <- file.path(Sys.getenv("HOME"), ".Rprofile")
REPOCODE <- paste0('options(repos = c(DRAN = "', DRANREPOURL, '"))\n')
DEPPKGS <- c("stringr", "git2r", "credentials", "httr")
# Ensure necessary packages are installed
install_cran <- function(pkgnames) {
  invisible(unlist(sapply(pkgnames, \(pkgname) {
    if (!requireNamespace(pkgname, quietly = TRUE)) {
      install.packages(pkgname, dependencies = TRUE)
    }
  })))
}
install_cran(DEPPKGS)
# try git

#' @export
has_git <- function(gitrooturl = GITROOTURL) {
  # Function to check if Git is available
  check_git_available <- function() {
    git_check <- tryCatch(
      {
        credentials::git_credential_ask(gitrooturl)
        TRUE
      },
      error = function(e) {
        FALSE
      }
    )
    return(git_check)
  }
  # Check if Git is available
  if (check_git_available()) {
    return(invisible(TRUE))
  }
  message("Git does not appear to be installed or is not working properly.")
  os_type <- .Platform$OS.type
  if (os_type == "windows") {
    message("It looks like you're using Windows.")
    message("To install Git on Windows, follow these steps:")
    message("1. Download the Git installer from https://git-scm.com/download/win")
    message("2. Run the installer and follow the on-screen instructions.")
    message("3. After installation, you may need to restart your computer or log out and back in.")
  } else if (os_type == "unix") {
    if (Sys.info()["sysname"] == "Darwin") {
      # macOS
      message("It looks like you're using macOS.")
      message("To install Git on macOS, follow these steps:")
      message("1. Open the Terminal.")
      message("2. Install Git using Homebrew with the following command:
        brew install git")
      message("3. If you don't have Homebrew installed,
        you can install it from https://brew.sh")
    } else if (Sys.info()["sysname"] == "Linux") {
      # Linux
      message("It looks like you're using Linux.")
      message("To install Git on Linux, follow these steps:")
      message("1. Open the Terminal.")
      message("2. Install Git using your package manager. For example:")
      message("   - On Debian-based systems (e.g., Ubuntu):
        sudo apt-get install git")
      message("   - On Red Hat-based systems (e.g., Fedora):
        sudo dnf install git")
      message("   - On Arch-based systems: sudo pacman -S git")
    } else {
      message("Your operating system is not recognized.
        Please refer to the Git installation guide at
        https://git-scm.com/book/en/v2/Getting-Started-Installing-Git")
    }
  } else {
    message("Your operating system is not recognized.
      Please refer to the Git installation guide at
      https://git-scm.com/book/en/v2/Getting-Started-Installing-Git")
  }
  return(invisible(FALSE))
}
# has_git()

#' Add a Custom Repository to .Rprofile
#'
#' This function appends a custom repository setting to the `.Rprofile` file if
#' it does not already exist. If the `.Rprofile` file does not exist, it creates
#' one and adds the repository setting.
#'
#' @param rprofilefn A character string specifying the path to the `.Rprofile`
#' file. Defaults to `RPROFILEFN`.
#' @param repocode A character string containing the R code to set the
#' repository. Defaults to `REPOCODE`.
#'
#' @details
#' The function checks if the `.Rprofile` file exists. If it does, it reads
#' the file and checks for the presence of existing repository settings. If
#' the repository settings are not found, it appends the new repository code.
#' If the `.Rprofile` file does not exist, it creates the file and writes the
#' repository settings to it.
#'
#' @note The default values for `rprofilefn` and `repocode` should be specified
#' in your environment or script where this function is used.
#'
#' @export
#'
#' @examples
#' # Add the custom repository to .Rprofile
#' drwater_repo()
drwater_repo <- function(rprofilefn = RPROFILEFN,
                         repocode = REPOCODE) {
  if (file.exists(rprofilefn)) {
    # Read the existing .Rprofile
    existing_code <- readLines(rprofilefn)
    # Check if the repository setting already exists
    repo_exists <- any(grepl("options\\(repos", existing_code))
    dran_exists <- any(grepl("DRAN", existing_code))
    if (!repo_exists) {
      # Append the code if no repository settings are present
      cat(repocode, file = rprofilefn, append = TRUE)
    } else if (!dran_exists) {
      # Warn if repository settings exist but DRAN is missing
      message(".Rprofile already contains repository settings.")
      message(
        "Please add the DRAN repo to your current ", rprofilefn,
        ", if it is not included!"
      )
      message("Modify to something like:")
      message("options(repos = c(CRAN = \"https://cloud.r-project.org\", DRAN = \"https://dran.drwater.net\"))")
    } else {
      # Inform that repository settings are already present
      message(".Rprofile already contains repository settings.")
    }
  } else {
    # Create the .Rprofile and add the code
    cat(repocode, file = rprofilefn)
  }
}


#' Download Package from drwateR Git Repository
#'
#' Downloads a package from the drwateR Git repository.
#'
#' @param pkgname Character; Name of the package to download.
#' @param gitroot Character; The root URL of the Git repository
#' (default: `DRWATERGITROOT`).
#' @param scheme Character; The URL scheme to use (default: "https").
#' @param has_git Logical; Whether git is installed and available (default: `has_git()`).
#'
#' @return Character; The local path of the downloaded package or NULL
#' if the download fails.
#'
#' @importFrom git2r clone
#' @importFrom credentials git_credential_ask
#'
#' @export
#'
#' @examples
#' download_drwateR("rmdify")
download_drwateR <- function(pkgname,
                             gitroot = DRWATERGITROOT,
                             scheme = DRWATERGITSCHEME) {
  gitrooturl <- paste0(scheme, "://", gitroot)
  has_git <- has_git(gitrooturl)
  cred <- credentials::git_credential_ask(gitrooturl)
  if (!has_git) {
    stop("`git` is not ready!")
  }
  # Request credentials
  # Construct the package URL
  pkgurl <- if (is.null(cred)) {
    paste0(gitrooturl, "/drwateR/", pkgname, ".git")
  } else {
    paste0(
      scheme, "://", cred$username, ":", cred$password, "@",
      gitroot, "/drwateR/", pkgname, ".git"
    )
  }
  # Define local path for the clone
  localpath <- file.path(tempdir(), paste0(pkgname, ".git"))
  # Remove existing directory if it exists
  if (dir.exists(localpath)) {
    unlink(localpath, recursive = TRUE)
  }
  # Clone the repository and return the path
  path <- tryCatch(
    {
      cloned_path <- git2r::clone(pkgurl, local_path = localpath)
      gsub("/\\.git$", "", cloned_path$path)
    },
    error = function(e) {
      return(NULL)
    }
  )
  return(path)
}


#' Install Package From drwateR DRAN or Git Repository
#'
#' Installs a package from the drwateR Git repository or the DRAN repository, including handling
#' dependencies.
#'
#' @param pkgname Character; Name of the package to install.
#' @param dependency_reinstall Logical; If TRUE, reinstall dependencies.
#' @param pkgpath Character; Local path to the downloaded package (default:
#' path returned by `download_drwateR`).
#' @param drwaterpkgs Character; A vector of package names belonging to
#' `drwateR`.
#' @param from Character; Source of the package, either "dran" or "git".
#' @param repos Character; URL of the DRAN repository.
#' @param force Logical; If TRUE, force the installation even if the
#' package is up-to-date.
#'
#' @return Logical; TRUE if the package was successfully installed,
#' FALSE otherwise.
#'
#' @importFrom utils install.packages
#' @export
#'
#' @examples
#' install_drwateR("rmdify")
install_drwateR <- function(pkgname,
                            dependency_reinstall = FALSE,
                            from = "dran",
                            drwaterpkgs = DRWATERPKGS,
                            core = CORE,
                            repos = DRANREPOURL,
                            force = FALSE) {
  pkgname <- intersect(pkgname, drwaterpkgs)
  if (length(pkgname) == 0) {
    warning("No packages from DRAN found!")
    return(invisible(FALSE))
  }
  if ("drwateR" %in% pkgname) {
    pkgname <- unique(core, pkgname)
  }
  pkgname <- get_drandenpend(pkgname)
  # Validate package name
  # install core packages
  invisible(sapply(pkgname, function(i) {
    # Install from specified source
    if (from == "git") {
      install_drwateR_git(i,
        dependency_reinstall = dependency_reinstall,
        force = force
      )
    } else {
      install_drwateR_dran(
        pkgname = i,
        dranrepo = repos
      )
    }
  }))
  return(invisible(TRUE))
}


#' Install a Package from Git Repository
#'
#' This function installs an R package from a specified Git repository.
#' It handles dependencies, checks for version updates, and performs the
#' installation.
#'
#' @param pkgname A character string specifying the name of the package to
#' be installed.
#' @param dependency_reinstall A logical value indicating whether to
#' reinstall dependencies
#'   if they are already installed. Default is `FALSE`.
#' @param pkgpath A character string specifying the path to the package
#' tarball. If not provided,
#'   it is determined by the `download_drwateR` function. Default is
#' `download_drwateR(pkgname)`.
#' @param force A logical value indicating whether to force the
#' installation even if the
#'   local version matches the remote version. Default is `FALSE`.
#' @return A logical value indicating whether the installation was
#' successful. Returns `TRUE`
#'   if the package was installed successfully, otherwise `FALSE`.
#' The return value is invisible.
#' @importFrom git2r clone
#' @importFrom credentials git_credential_ask
#' @export
install_drwateR_git <- function(
    pkgname,
    dependency_reinstall = FALSE,
    pkgpath = download_drwateR(pkgname),
    force = FALSE) {
  # Validate the package path
  if (is.null(pkgpath) || !file.exists(pkgpath)) {
    stop(pkgname, " has not been successfully downloaded!")
  }
  # Handle dependencies
  depdf <- dependency_drwateR(pkgpath)
  if (nrow(depdf) > 0) {
    install_dependency_git(depdf, dependency_reinstall = dependency_reinstall, force = force)
  }
  # Check for version updates
  if (requireNamespace(pkgname, quietly = TRUE)) {
    remoteversion <- unname(read.dcf(file.path(
      pkgpath,
      "DESCRIPTION"
    ))[, "Version"])
    localversion <- as.character(packageVersion(pkgname))
    if (remoteversion == localversion && !force) {
      message("No update in remote!")
      unlink(pkgpath, recursive = TRUE, force = FALSE, expand = TRUE)
      return(invisible(FALSE))
    }
  }
  # Install the package
  result <- tryCatch(
    system(paste("R CMD INSTALL", shQuote(pkgpath)), intern = TRUE),
    error = function(e) {
      warning("Failed to install package from ", pkgpath, ": ", e$message)
      return(NULL)
    }
  )
  if (is.null(result)) {
    warning("R CMD INSTALL exited with status ", result)
    out <- FALSE
  } else {
    invisible(sapply(result, message))
    message("Package installed successfully from ", pkgpath)
    out <- TRUE
  }
  unlink(pkgpath, recursive = TRUE, force = TRUE)
  return(invisible(out))
}

#' Install dependencies for a package
#'
#' Installs the dependencies for a specified package from various sources
#' including CRAN, GitHub, and drwateR.
#'
#' @param depdf A data frame containing information about dependency packages.
#' Must include columns `pkgname`, `source`, and `installed`.
#' @param dependency_reinstall Logical; if TRUE, reinstall all dependencies
#' regardless of their installation status.
#'
#' @importFrom devtools install_github
#' @importFrom utils install.packages
#' @export
#'
install_dependency_git <- function(depdf, dependency_reinstall = FALSE, force = FALSE) {
  # Filter installed packages based on the `dependency_reinstall` flag
  if (!dependency_reinstall) {
    depdf <- depdf[!depdf$installed, ]
  }
  # Install CRAN packages
  cranpkgs <- depdf[depdf$source == "cran", "pkgname"]
  if (length(cranpkgs) > 0) {
    message("Installing CRAN packages: ", paste(cranpkgs, collapse = ", "))
    install.packages(cranpkgs)
  }
  # Install GitHub packages
  githubpkgs <- depdf[depdf$source == "github", "pkgname"]
  if (length(githubpkgs) > 0) {
    message("Installing GitHub packages: ", paste(githubpkgs, collapse = ", "))
    invisible(sapply(githubpkgs, function(pkg) {
      suppressMessages(devtools::install_github(pkg))
    }))
  }
  # Install drwateR packages
  drwateRpkgs <- depdf[depdf$source == "drwateR", "pkgname"]
  if (length(drwateRpkgs) > 0) {
    message(
      "Installing drwateR packages: ",
      paste(drwateRpkgs, collapse = ", ")
    )
    invisible(sapply(drwateRpkgs, \(x) {
      install_drwateR_git(
        pkgname = x,
        dependency_reinstall = dependency_reinstall,
        force = force
      )
    }))
  }
}


#' Find all dependency packages for a given package
#'
#' This function identifies all dependency packages listed in the
#' DESCRIPTION file of a local R package source code. It categorizes
#' the dependencies based on their source: CRAN, drwateR, or GitHub.
#'
#' @param pkgpath Character; the local path to the package source code.
#' @param desc A data frame containing the content of the DESCRIPTION file.
#'              If not provided, it will be read from the specified `pkgpath`.
#' @param drwaterpkgs Character vector; the list of drwateR packages.
#'
#' @return A data.frame containing the names of dependency packages,
#'         whether they are installed, and their source.
#'
#' @export
#'
dependency_drwateR <- function(
    pkgpath,
    desc = read.dcf(file.path(pkgpath, "DESCRIPTION")),
    drwaterpkgs = DRWATERPKGS) {
  # Extract dependencies from DESCRIPTION
  depends <- if ("Depends" %in% colnames(desc)) {
    gsub(" (.*)$", "", unlist(strsplit(unname(desc[, "Depends"]), split = ",\\n")))
  } else {
    character(0)
  }
  imports <- if ("Imports" %in% colnames(desc)) {
    gsub(" (.*)$", "", unlist(strsplit(unname(desc[, "Imports"]), split = ",\\n")))
  } else {
    character(0)
  }
  remotes <- if ("Remotes" %in% colnames(desc)) {
    gsub(" (.*)$", "", unlist(strsplit(unname(desc[, "Remotes"]), split = ",\\n")))
  } else {
    character(0)
  }
  # Combine unique package names
  pkgname <- unique(c(depends, imports, remotes))
  pkgname <- pkgname[pkgname != "R"]
  # Create data frame of dependencies
  depdf <- data.frame(
    pkgname = pkgname,
    installed = sapply(
      pkgname,
      function(x) requireNamespace(x, quietly = TRUE)
    ),
    stringsAsFactors = FALSE
  )
  # Assign sources
  depdf$source <- "cran"
  depdf$source[depdf$pkgname %in% drwaterpkgs] <- "drwateR"
  depdf$source[grepl("\\/", depdf$pkgname)] <- "github"
  return(depdf)
}

#' Update drwateR packages
#'
#' Updates one or more packages from the drwateR Git repository.
#'
#' @param pkgname Character vector; the names of packages to update.
#'                If `NULL`, updates all drwateR packages by default.
#'
#' @return Logical; `TRUE` if the packages were successfully updated,
#'         `FALSE` otherwise.
#' @export
#'
#' @examples
#' update_drwateR()
#'
update_drwateR <- function(pkgname = DRWATERPKGS, ...) {
  # Install each package and check for success
  results <- sapply(pkgname, function(pkg) {
    tryCatch(
      {
        install_drwateR(pkg, ...)
        TRUE # Return TRUE if installation succeeds
      },
      error = function(e) {
        warning(paste("Failed to update package:", pkg, "-", e$message))
        FALSE # Return FALSE if installation fails
      }
    )
  })
  # Return TRUE if all packages were successfully updated, FALSE otherwise
  return(invisible(all(results)))
}


# --------------------------------------------------------------------


# Function to fetch credentials from .netrc
fetch_credentials_from_netrc <- function(url) {
  # Determine the appropriate file path for the .netrc file
  netrc_file <- file.path(Sys.getenv("HOME"), ".netrc")
  if (!file.exists(netrc_file)) {
    return(prompt_and_store_credentials(url))
  }
  # Read the .netrc file
  netrc_content <- readLines(netrc_file)
  # Extract the machine name from the URL
  machine <- stringr::str_extract(url, "https?://([^/]+)") |>
    stringr::str_remove("https?://")
  # Parse the .netrc file to find the credentials for the specified machine
  machine_index <- grep(paste0("^machine\\s+", machine), netrc_content)
  if (length(machine_index) == 0) {
    return(prompt_and_store_credentials(url))
  }
  # Initialize variables
  username <- NULL
  password <- NULL
  # Extract login and password
  for (i in seq(machine_index + 1, length(netrc_content))) {
    if (grepl("^\\s*login\\s+", netrc_content[i])) {
      username <- stringr::str_trim(sub("^\\s*login\\s+", "", netrc_content[i]))
    }
    if (grepl("^\\s*password\\s+", netrc_content[i])) {
      password <- stringr::str_trim(sub(
        "^\\s*password\\s+", "",
        netrc_content[i]
      ))
    }
    if (!is.null(username) && !is.null(password)) {
      return(list(username = username, password = password))
    }
  }
  return(prompt_and_store_credentials(url))
}

# Function to validate the credentials (simple validation,
# can be adjusted as needed)
validate_credentials <- function(username, password) {
  # Simple validation: Ensure neither username nor password is empty
  return(nzchar(username) && nzchar(password))
}

# Function to prompt the user for credentials and store them in .netrc,
# with validation and retry logic
prompt_and_store_credentials <- function(url) {
  max_attempts <- 3
  attempts <- 0
  while (attempts < max_attempts) {
    # Prompt the user for username and password
    username <- readline(prompt = paste0(
      "Please input the credentials for '",
      url, "'\n\nEnter username: "
    ))
    password <- readline(prompt = "Enter password: ")
    # Validate the credentials
    if (validate_credentials(username, password)) {
      # Determine the appropriate file path for the .netrc file
      netrc_file <- file.path(Sys.getenv("HOME"), ".netrc")
      # Extract the machine name from the URL
      machine <- stringr::str_extract(url, "https?://([^/]+)") |>
        stringr::str_remove("https?://")
      # Create the entry for the .netrc file
      entry <- paste0(
        "machine ", machine,
        "\n  login ", username,
        "\n  password ", password, "\n"
      )
      # Append the entry to the .netrc file, creating the
      # file if it doesn't exist
      write(entry, file = netrc_file, append = TRUE)
      # Return the credentials
      return(list(username = username, password = password))
    } else {
      cat("Invalid username or password. Please try again.\n")
      attempts <- attempts + 1
    }
  }
  cat("Exceeded maximum attempts. Returning NULL.\n")
  return(NULL)
}





#' Download a file with credentials
#'
#' @param url The URL to download from.
#' @param destdir The directory to save the downloaded file.
#' @param destfn The name of the file to save.
#' @param credentials A list with username and password for authentication.
#'
#' @importFrom httr GET authenticate content
#'
#'
download_with_credentials <- function(url, destdir, destfn, credentials) {
  destfile <- file.path(destdir, destfn)
  response <- httr::GET(url, httr::authenticate(credentials$username, credentials$password))
  binary_content <- httr::content(response, "raw")
  writeBin(binary_content, destfile)
}

#' Fetch package information from a CRAN-like repository
#'
#' @param pkgname The name of the package to fetch information for.
#' @param dranrepo The URL of the CRAN-like repository. Default is "https://drwater.rcees.ac.cn/dran".
#' @return A list with package information including tarball link, dependencies, and imports.
#' @export
#'
drwater_pkginfo <- function(
    pkgname,
    dranrepo = DRANREPOURL,
    pkgtype = PACKAGETYPE) {
  # Download the PACKAGES file to a temporary location
  pkg_info <- data.frame()
  for (i in seq_along(pkgtype)) {
    packages_url <- file.path(dranrepo, pkgtype[i], "src/contrib/PACKAGES")
    tmpfile <- tempfile()
    invisible(download_with_credentials(packages_url,
      destdir = dirname(tmpfile),
      destfn = basename(tmpfile),
      credentials = fetch_credentials_from_netrc(packages_url)
    ))
    packages_info <- read.dcf(tmpfile)
    # Search for the package information
    info <- as.data.frame(
      t(packages_info[packages_info[, "Package"] == pkgname, ])
    )
    if (ncol(info) > 0) {
      info$type <- pkgtype[i]
    }
    pkg_info <- rbind(pkg_info, info)
  }
  if (ncol(pkg_info) == 0) {
    stop(paste("Package", pkgname, "not found in the PACKAGES file."))
  }
  # Extract relevant details
  pkg_version <- pkg_info$Version
  pkg_deps <- pkg_info$Depends
  pkg_imports <- pkg_info$Imports
  pkg_remotes <- pkg_info$Remotes
  pkg_type <- pkg_info$type
  # Construct the download link for the package tarball
  pkg_tarball <- paste0(
    dranrepo, "/", pkg_type, "/src/contrib/",
    pkgname, "_", pkg_version, ".tar.gz"
  )
  return(list(
    package = pkgname,
    version = pkg_version,
    dependencies = pkg_deps,
    imports = pkg_imports,
    remotes = pkg_remotes,
    type = pkg_type,
    tarball_link = pkg_tarball
  ))
}




get_dranpkginfo <- function(dranrepo = DRANREPOURL) {
  packages_url <- file.path(dranrepo, "src/contrib/PACKAGES")
  # Download the PACKAGES file to a temporary location
  tmpfile <- tempfile()
  invisible(download_with_credentials(packages_url,
    destdir = dirname(tmpfile),
    destfn = basename(tmpfile),
    credentials = fetch_credentials_from_netrc(packages_url)
  ))
  dranpkginfo <- read.dcf(tmpfile)
  return(dranpkginfo)
}

install_dran <- function(
    pkgname,
    dranrepo = DRANREPOURL,
    pkginfo = drwater_pkginfo(pkgname, dranrepo = dranrepo),
    credentials = fetch_credentials_from_netrc(dranrepo)) {
  if (length(pkgname) > 1) {
    return(sapply(
      pkgname,
      \(x) {
        install_dran(x, dranrepo, pkginfo, credentials)
      }
    ))
  }
  if (requireNamespace(pkgname, quietly = TRUE)) {
    remoteversion <- pkginfo$version
    localversion <- gsub("^‘|’$", "", packageVersion(pkgname))
    if (localversion == remoteversion) {
      message("No reomote update for `", pkgname, "!")
      return(invisible(NULL))
    }
  }
  if (is.null(credentials)) {
    stop("Credentials not found in .netrc file.")
  }
  tmpfile <- tempfile(fileext = ".tar.gz")
  download_with_credentials(
    pkginfo$tarball_link,
    dirname(tmpfile),
    basename(tmpfile),
    credentials
  )
  install.packages(tmpfile, repos = NULL, type = "source")
  unlink(tmpfile)
}

#' Install a package from drwater repository
#'
#' @param pkgname The name of the package to install.
#' @param dranrepo The URL of the CRAN-like repository.
#' Default is "https://drwater.rcees.ac.cn/dran".
#'
#' @importFrom stringr str_split str_trim str_remove str_extract
#'
#' @export
install_drwateR_dran <- function(
    pkgname,
    dranrepo = DRANREPOURL,
    dranpkgnames = get_dranpkginfo()[, "Package"],
    credentials = fetch_credentials_from_netrc(dranrepo)) {
  if (length(pkgname) > 1) {
    return(unlist(sapply(pkgname, \(x) install_drwateR_dran(x))))
  }
  pkginfo <- drwater_pkginfo(pkgname, dranrepo = dranrepo)
  deppkgs <- c(
    pkginfo$dependencies,
    unlist(strsplit(pkginfo$imports, ", |,\\n"))
  )
  deppkgs <- gsub(" (.*)$", "", deppkgs)
  deppkgs <- deppkgs[!is.na(deppkgs)]
  deppkgs <- deppkgs[deppkgs != "R"]
  cranpkgs <- setdiff(deppkgs, dranpkgnames)
  install_cran(cranpkgs)
  if (!is.null(pkginfo$remotes) && !is.na(pkginfo$remotes)) {
    githubpkgs <- unlist(strsplit(pkginfo$remotes, ", |,\\n"))
    githubpkgs <- githubpkgs[!is.na(githubpkgs)]
    if (length(githubpkgs) > 0) {
      invisible(sapply(githubpkgs, devtools::install_github))
    }
  }
  install_dran(
    pkgname = pkgname,
    dranrepo = dranrepo,
    pkginfo = drwater_pkginfo(pkgname, dranrepo = dranrepo),
    credentials = credentials
  )
  return(invisible(TRUE))
}

# 拓扑排序帮助函数
topological_sort <- function(dependency_graph) {
  sorted_packages <- c()
  temp_mark <- c()

  visit <- function(node) {
    if (node %in% sorted_packages) {
      return()
    }
    if (node %in% temp_mark) stop("Circular dependency detected")

    temp_mark <<- c(temp_mark, node)
    if (!is.null(dependency_graph[[node]])) {
      for (dependency in dependency_graph[[node]]) {
        visit(dependency)
      }
    }
    temp_mark <<- temp_mark[temp_mark != node]
    sorted_packages <<- c(node, sorted_packages)
  }

  for (node in names(dependency_graph)) {
    if (!(node %in% sorted_packages)) {
      visit(node)
    }
  }

  return(sorted_packages)
}

get_drandenpend <- function(pkgname, pkginfo = get_dranpkginfo()) {
  if (length(pkgname) > 1) {
    drandenpend <- unique(unlist(lapply(
      pkgname,
      \(x) get_drandenpend(x, pkginfo)
    )))
    return(drandenpend)
  }
  # 读取并解析 PACKAGES 二进制内容
  # 提取包名和依赖项
  dranpkgnames <- pkginfo[, "Package"]
  depends <- pkginfo[, "Depends"]
  imports <- pkginfo[, "Imports"]
  # 合并依赖项和导入项
  dependencies <- paste(depends, imports, sep = ",")
  dependencies <- strsplit(dependencies, ",")
  names(dependencies) <- dranpkgnames
  # 定义递归函数来获取所有依赖项
  get_all_dependencies <- function(pkg,
                                   dependencies,
                                   dr_packages,
                                   seen = character()) {
    if (pkg %in% seen) {
      return(NULL)
    } # 防止循环依赖
    seen <- c(seen, pkg)
    direct_deps <- trimws(dependencies[[pkg]])
    direct_deps <- intersect(direct_deps, dr_packages) # 只保留 DRAN 内部的包
    all_deps <- unique(direct_deps)
    for (dep in direct_deps) {
      sub_deps <- get_all_dependencies(dep, dependencies, dr_packages, seen)
      all_deps <- unique(c(all_deps, sub_deps))
    }
    return(all_deps)
  }
  # 获取指定包的所有依赖项
  if (!(pkgname %in% dranpkgnames)) {
    stop("The specified package is not found in the DRAN repository.")
  }
  drandenpend <- rev(get_all_dependencies(pkgname, dependencies, dranpkgnames))
  # 创建依赖图
  dependency_graph <- setNames(vector(
    "list",
    length(dranpkgnames)
  ), dranpkgnames)
  for (pkg in dranpkgnames) {
    dependency_graph[[pkg]] <- get_all_dependencies(
      pkg,
      dependencies,
      dranpkgnames
    )
  }
  # 对依赖项进行拓扑排序
  sorted_dependencies <- topological_sort(dependency_graph)
  # 过滤只包含指定包及其依赖项的排序结果
  sorted_drandenpend <- rev(intersect(
    sorted_dependencies,
    c(pkgname, drandenpend)
  ))
  return(sorted_drandenpend)
}
