#' Install packages from a Git repository
#'
#' Downloads and installs an R package from a Git repository.
#'
#' @param pkgpath The GitHub repository path in the format "user/repo".
#' @param from The base URL of the Git repository (default: "https://hub.fastgit.org/").
#'
#' @return Logical value indicating if the installation was successful.
#' @export
#'
#' @examples
#' install_giturl("tidyverse/tidyverse")
install_giturl <- function(pkgpath,
                           from = paste0("https://", DRWATERGITROOT)) {
  if (is.null(pkgpath) || !grepl("\\/", pkgpath)) {
    stop("Valid package path required in the format 'user/repo'")
  }
  pkgurl <- paste0(from, pkgpath, ".git")
  pkgname <- gsub("^.*\\/", "", pkgpath)
  # Clone the repository
  local_path <- file.path(tempdir(), paste0(pkgname, ".git"))
  tryCatch(
    git2r::clone(pkgurl, local_path = local_path),
    error = function(e) {
      stop("Failed to clone the repository: ", e$message)
    }
  )
  # Install the package
  install_cmd <- paste("R CMD INSTALL", local_path)
  install_status <- system(install_cmd,
    intern = TRUE,
    ignore.stdout = TRUE,
    ignore.stderr = TRUE
  )
  # Clean up
  unlink(local_path, recursive = TRUE, force = TRUE)
  if (install_status != 0) {
    stop("Failed to install the package.")
  }
  return(invisible(TRUE))
}
